# -*- coding: utf-8 -*-
# Author: Aethlios <tom.chambaretaud@protonmail.com>

from typing import Callable, Any, Optional
from functools import partial
from decimal import Decimal
from math import floor

from resetTolkien.format import Formatter, FormatType, EncodingType, HashingType
from resetTolkien.utils import (
    NotAHash,
    GeneratorLen,
    AlternativeGen,
    TimestampGenerator,
    TimestampHashFormat,
    import_from_yaml,
    possible_date_format_of_token,
    CustomFloat,
)
from resetTolkien.constants import (
    DEFAULT_TIMERANGE_FOR_INT_TIMESTAMP,
    DEFAULT_TIMERANGE_FOR_FLOAT_TIMESTAMP,
    MAX_DEPTH_LEVEL,
    DEFAULT_THREAD_NUMBER,
    DEFAULT_CONFIG_FILE,
    DEFAULT_DECIMAL_LENGTH,
)


class ResetTolkien:
    """Class to detect whether a value is based on a timestamp
    and generate tokens with the correct formatting."""

    def __init__(
        self,
        token: Optional[str] = None,
        verbosity: int = 0,
        timestamp_hash_formats_config_file: str = DEFAULT_CONFIG_FILE,
        level: int = MAX_DEPTH_LEVEL,
        decimal_length: int = DEFAULT_DECIMAL_LENGTH,
        int_range_limit: int = DEFAULT_TIMERANGE_FOR_INT_TIMESTAMP,
        float_range_limit: int = DEFAULT_TIMERANGE_FOR_FLOAT_TIMESTAMP,
        prefixes: Optional[list[str]] = None,
        suffixes: Optional[list[str]] = None,
        hashes: Optional[list[str]] = None,
        timezone: int = 0,
        date_format_of_token: Optional[str] = None,
        formats: Optional[list[str]] = None,
    ) -> None:
        """Initialization function of ResetTolkien."""

        self.formatter = Formatter()
        self.token = token
        self.decimal_length = (
            decimal_length if decimal_length else DEFAULT_DECIMAL_LENGTH
        )
        self.int_range_limit = (
            int_range_limit * 2
            if int_range_limit
            else DEFAULT_TIMERANGE_FOR_INT_TIMESTAMP * 2
        )
        self.float_range_limit = (
            float_range_limit * 2 * 10**self.decimal_length
            if float_range_limit
            else DEFAULT_TIMERANGE_FOR_FLOAT_TIMESTAMP * 2 * 10**self.decimal_length
        )
        self.prefixes: list[str] = [] if not prefixes else prefixes.copy()
        self.suffixes: list[str] = [] if not suffixes else suffixes.copy()
        self.hashes: list[str] = []
        if hashes:
            for h in hashes:
                if not h in self.formatter.allHashing():
                    raise ValueError(f"Unknown hash : {h}")
            self.hashes = hashes.copy()
        self.timezone = timezone
        self.date_format_of_token = date_format_of_token
        self.formats = []
        if formats:
            self.formats = self.formatter.import_formats(formats)
        self.verbosity = verbosity
        self.level = level
        self.timestamp_hash_formats_config_file = timestamp_hash_formats_config_file
        self.timestamp_hash_formats = self.getTimestampsHashFormatsByLevel(
            level, timestamp_hash_formats_config_file
        )

    def getTimestampsHashFormatsByLevel(
        self, level: int, timestamp_hash_formats_config_file: str
    ) -> list[TimestampHashFormat]:
        """Class function to get depth level of timestamp hash search."""

        if level > MAX_DEPTH_LEVEL or level < 1:
            raise ValueError

        if self.verbosity >= 1:
            print(f"Selected level : {level}")

        timestamp_hash_formats = import_from_yaml(
            timestamp_hash_formats_config_file,
            level,
            self.int_range_limit,
            self.float_range_limit,
            self.formatter.intHashing(),
            self.formatter.floatHashing(),
            self.formatter.allEncoding(),
        )

        if self.verbosity >= 1:
            print(f"Selected Timestamp hash formats : {timestamp_hash_formats}")

        return timestamp_hash_formats

    def detectOneEncoding(
        self, token: str, format: EncodingType, **kwargs: Any
    ) -> tuple[Optional[str], Optional[EncodingType]]:
        """Determines whether a value has been encoded
        using an encoding function provided as input."""

        try:
            c = format(token, encode=False, **kwargs)
            return c, format
        except Exception as e:
            if self.verbosity >= 2:
                print(f"Exception with {format.__name__} : {e}")
            return None, None

    def detectOneHash(
        self,
        token: str,
        hash: HashingType,
        possibleTokens: GeneratorLen,
        multithreading: Optional[int] = None,
    ) -> tuple[
        tuple[Optional[str], Optional[str], Optional[str]], Optional[HashingType]
    ]:
        """Determines whether a value has been hashed
        using an hash function provided as input."""

        try:
            c = hash(
                token,
                possibleTokens,
                encode=False,
                multithreading=multithreading,
                prefixes=self.prefixes,
                suffixes=self.suffixes,
            )
            if isinstance(c, tuple):
                return c, hash
            else:
                raise ValueError("The token has not been decoded but encoded")
        except NotAHash as e:
            if self.verbosity >= 2:
                print(f"Exception with {hash.__name__} : {e}")
            return (None, None, None), None

    def detectHash(
        self,
        hashes: dict[str, HashingType],
        hashes_by_type: dict[str, HashingType],
        formats: list[FormatType],
        formats_output: list[FormatType],
        timestamp: float,
        token: str,
        timestamp_type_func: Callable[[str], Any],
        range_limit: int,
        multithreading: Optional[int] = None,
    ) -> Optional[
        tuple[tuple[str, Optional[str], Optional[str]], list[FormatType], bool]
    ]:
        """Determines which hash function matches the format of an input token."""

        for h in hashes:
            if h in hashes_by_type and (len(self.hashes) == 0 or h in self.hashes):
                hash = hashes[h]
                possibleTokens = GeneratorLen(
                    self.generate_possible_token(
                        timestamp_type_func(str(timestamp)),
                        range_limit=range_limit,
                        formats=formats_output,
                    ),
                    range_limit,
                )
                (new_token, prefix, suffix), format = self.detectOneHash(
                    token, hash, possibleTokens, multithreading=multithreading
                )
                if new_token and format:
                    if self.verbosity >= 1:
                        print(f"Hash found! : {h}")
                    new_formats: list[FormatType] = formats.copy()
                    new_formats.append(format)
                    if new_token[1] or new_token[2]:
                        new_formats.append(
                            partial(
                                self.formatter.prefix_suffix,
                                prefix=prefix,
                            )
                        )
                        new_formats.append(
                            partial(
                                self.formatter.prefix_suffix,
                                suffix=suffix,
                            )
                        )
                    new_formats += formats_output
                    return ((new_token, prefix, suffix), new_formats, True)
        return None

    def _detectFormat(
        self,
        token: str,
        multithreading: int,
        formats: Optional[list[FormatType]] = None,
        timestamp: Optional[float] = None,
    ) -> list[tuple[tuple[str, Optional[str], Optional[str]], list[FormatType], bool]]:
        """Recursive function - Determines a function
        corresponding to the format of an input token."""

        formats = [] if formats is None else formats

        numbers = self.formatter.getNumbers(token)
        if self.verbosity >= 1 and len(numbers) > 0:
            print(f"Integer value detected : {numbers}")
        possibleTimestamp = self.formatter.searchTimestamps(numbers)
        if len(possibleTimestamp) > 0:
            print(f"Possible timestamp detected! {possibleTimestamp} from \"{token}\"")

        if self.formatter.isLiteralIntegerOrFloat(token):
            new_token, _ = self.detectOneEncoding(token, self.formatter.timestamp)
            if new_token:
                if self.verbosity >= 1:
                    print(
                        f"Format detected and stored : {self.formatter.export_formats(formats)}"
                    )
                self.formats = formats
                return [((token, None, None), formats, True)]

        if self.formatter.isDatetime(token):
            if self.verbosity >= 1:
                print(f"Datetime value detected : {token}")
            new_token, format = self.detectOneEncoding(
                token,
                self.formatter.datetime,
                timezone=self.timezone,
                date_format_of_token=self.date_format_of_token,
            )
            if new_token and format:
                if not self.date_format_of_token:
                    possible_format = possible_date_format_of_token(token)
                    print(
                        f'Please enter a datetime format for this input: "{token}" (possible format : --date-format-of-token="{possible_format}")'
                    )
                    exit()
                formats.append(format)
                if self.verbosity >= 1:
                    print(
                        f"Format detected and stored : {self.formatter.export_formats(formats)}"
                    )
                self.formats = formats
                return [((token, None, None), formats, True)]

        results: list[
            tuple[tuple[str, Optional[str], Optional[str]], list[FormatType], bool]
        ] = []
        encodings = self.formatter.availableEncoding()
        for e in encodings:
            encoding = encodings[e]
            if len(formats) < 3 or formats[-3:] != [encoding, encoding, encoding]:
                if self.verbosity >= 2:
                    print(f"Check encoding with {e}")
                new_token, format = self.detectOneEncoding(
                    token, encoding, init_token=self.token
                )
                if new_token and format:
                    if self.verbosity >= 1:
                        print(f"Possible encoding format found : {format.__name__}")
                    new_formats = formats.copy()
                    new_formats.append(format)
                    results += self._detectFormat(
                        new_token,
                        multithreading,
                        formats=new_formats,
                        timestamp=timestamp,
                    )

        if timestamp:
            hashes = self.formatter.availableHashingDict(token)
            if len(hashes) > 0:
                if self.verbosity >= 1:
                    print(f"Hash format detected : {list(hashes.keys())} : {token}")

                if token.lower() != token:
                    if self.verbosity >= 1:
                        print(
                            f"Hash with uppercase detected : {token} -> {token.lower()}"
                        )
                    token = token.lower()

                for timestampFormat in self.timestamp_hash_formats:
                    if self.verbosity >= 1:
                        print(
                            f"Check hash with {timestampFormat.description} ({timestampFormat.range_limit} tokens)"
                        )

                    r = self.detectHash(
                        hashes,
                        timestampFormat.hashes_by_type,
                        formats,
                        timestampFormat.formats_output,
                        timestamp,
                        token,
                        timestampFormat.timestamp_type_func,
                        timestampFormat.range_limit,
                        multithreading=multithreading,
                    )
                    if r:
                        results.append(r)
                        break

        return results

    def detectFormat(
        self,
        timestamp: Optional[float] = None,
        nb_threads: int = DEFAULT_THREAD_NUMBER,
    ) -> Optional[
        list[tuple[tuple[str, Optional[str], Optional[str]], list[FormatType], bool]]
    ]:
        """Non-recursive function - Determines a list of functions
        corresponding to the format of an input token."""

        if not self.token:
            return None
        return self._detectFormat(self.token, nb_threads, timestamp=timestamp)

    def encode(self, value: str, formats: list[FormatType] | None = None) -> str:
        """Converts a value from an input list of format functions."""

        if formats == None:
            formats = self.formats

        for format in formats[::-1]:
            token = format(
                value,
                encode=True,
                timezone=self.timezone,
                date_format_of_token=self.date_format_of_token,
                init_token=self.token,
            )
            if isinstance(token, str):
                value = token
            else:
                raise ValueError("The token has not been encoded but decoded")

        return value

    def generate_possible_token(
        self,
        init: float,
        range_limit: int | None = None,
        formats: list[FormatType] | None = None,
        prefix: str | None = None,
        suffix: str | None = None,
    ) -> TimestampGenerator:
        """Returns a list of tokens generated from timestamps
        according to the value provided as input and
        the token generation format, taking into account
        the type of value provided."""

        if isinstance(init, int):
            if not range_limit:
                range_limit = self.int_range_limit
        else:
            if not range_limit:
                range_limit = self.float_range_limit

        for i in AlternativeGen(range_limit):  # type: ignore
            timestamp = str(init + i)
            if not isinstance(init, int):
                t = CustomFloat(init, self.decimal_length)
                t.value = t.value + i
                timestamp = str(t)
            if prefix:
                timestamp = f"{prefix}{timestamp}"
            if suffix:
                timestamp = f"{timestamp}{suffix}"
            encoded_timestamp = self.encode(timestamp, formats=formats)
            yield encoded_timestamp, timestamp

    def generate_bounded_possible_token(
        self,
        begin: float | int,
        end: float | int,
        formats: list[FormatType] | None = None,
        prefix: str | None = None,
        suffix: str | None = None,
    ) -> TimestampGenerator:
        """Returns a list of tokens generated from timestamps
        contained in a temporal range provided as input and
        the token generation format, taking into account
        the type of value provided.
        """

        if isinstance(begin, int) and isinstance(end, int):
            limit = end - begin + 1
            begin = end - floor((end - begin) / 2)
            return self.generate_possible_token(
                begin, prefix=prefix, suffix=suffix, range_limit=limit, formats=formats
            )
        else:
            limit = (
                int(
                    round(float(Decimal(end) - Decimal(begin)), self.decimal_length)
                    * 10**self.decimal_length
                )
                + 1
            )
            begin = round(
                (
                    float(
                        Decimal(end)
                        - Decimal(floor((limit - 1) / 2) / 10**self.decimal_length)
                    )
                ),
                self.decimal_length,
            )
            return self.generate_possible_token(
                begin, prefix=prefix, suffix=suffix, range_limit=limit, formats=formats
            )



# -*- coding: utf-8 -*-
# Author: Aethlios <tom.chambaretaud@protonmail.com>

import base64
import re
import concurrent.futures

from functools import partial
from typing import Callable, Any, Optional

import shortuuid
import uuid

from resetTolkien.utils import (
    FormatType,
    EncodingType,
    HashingType,
    HashIdentifierType,
    NotAHash,
    GeneratorLen,
    to_microsecond_timestamp,
    from_microsecond_timestamp,
    uniqid,
    deuniqid,
    detect_datetime,
    to_uuidv1,
    from_uuidv1,
    to_mongodb_objectid,
    from_mongodb_objectid,
    urlencode,
    urldecode,
    import_format_with_args,
)
from resetTolkien.constants import (
    DEFAULT_THREAD_NUMBER,
    DATE_SLASH_REGEX,
    DATE_DASH_REGEX,
    TIME_REGEX,
    MONTH_REGEX,
    PARTIAL_FUNC_NAME_REGEX,
    TIMESTAMP_STR_LENGTH,
)
from resetTolkien.hashes import Hashes


class Formatter:
    """Class for encoding, hashing and decrypting a timestamp-based value"""

    ### Utils
    def import_formats(self, formats: list[str]) -> list[FormatType]:
        """Converts a list of strings into a list of corresponding encoding and hash functions."""

        results: list[FormatType] = []
        format_funcs: dict[str, EncodingType | HashingType] = self.allFormat()

        for func_name in formats:
            if (
                "(" in func_name
                and func_name.split("(")[0] in self.allEncodingWithArgs()
            ):
                func, args = import_format_with_args(
                    func_name, self.allEncodingWithArgs()
                )
                format_func = partial(func, **args)
                results.append(format_func)
            else:
                if not func_name in format_funcs:
                    raise ValueError(f"{func_name} is not in available formats")
                results.append(format_funcs[func_name])
        return results

    def export_formats(self, formats: list[FormatType]) -> str:
        """Converts a list of encoding and hash functions into a list of strings."""

        formats_name: list[str] = []
        for format in formats:
            if isinstance(format, partial):
                elems = re.search(PARTIAL_FUNC_NAME_REGEX, str(format))
                if not elems:
                    raise ValueError(f"No partial name found : {format}")
                args = [
                    elem
                    for elem in elems.groups()[1].split(",")
                    if elem.split("=")[1] != "None"
                ]
                for arg in args:
                    name = f"{elems.groups()[0]}({arg})"
                    formats_name.append(name)
            else:
                formats_name.append(format.__name__)
        return ",".join(formats_name)

    def isLiteralIntegerOrFloat(self, token: str) -> bool:
        """Confirms that the value is numeric."""

        try:
            float(token)
            return True
        except (ValueError, OverflowError, TypeError):
            return False

    def getNumbers(self, token: str) -> list[str]:
        """Extracts numeric values from the string value."""

        matches = re.findall(r"([0-9]+\.?[0-9]*)", str(token))
        return [match for match in matches if self.isLiteralIntegerOrFloat(match)]

    def searchTimestamps(self, numbers: list[str]) -> list[float]:
        """Extracts nearby timestamp values from numerical values."""

        matches: list[float] = []
        for number in numbers:
            if self.isLiteralIntegerOrFloat(number):
                if "." in number:
                    n = float(number)
                elif len(number) > TIMESTAMP_STR_LENGTH:
                    n = float(
                        number[:TIMESTAMP_STR_LENGTH]
                        + "."
                        + number[TIMESTAMP_STR_LENGTH:]
                    )
                else:
                    n = int(number)
                try:
                    from_microsecond_timestamp(n)
                except ValueError:
                    continue
                matches.append(n)
        return matches

    def isDatetime(self, token: str) -> bool:
        """Confirms that the value is datetime format."""

        return (
            re.search(DATE_SLASH_REGEX, token) != None
            or re.search(DATE_DASH_REGEX, token) != None
            or re.search(TIME_REGEX, token) != None
            or re.search(MONTH_REGEX, token) != None
        )

    ### Format

    def allFormat(self) -> dict[str, EncodingType | HashingType]:
        """Returns all functions."""

        formats: dict[str, EncodingType | HashingType] = {}
        formats.update(self.allEncoding())
        formats.update(self.allHashing())
        return formats

    ### Encoding

    def allEncoding(self) -> dict[str, EncodingType]:
        """Returns all encoding functions."""

        return {
            "prefix_suffix": self.prefix_suffix,
            "base32": self.base32,
            "base64": self.base64,
            "urlencode": self.urlencode,
            "hexint": self.hexint,
            "hexstr": self.hexstr,
            "uniqid": self.uniqid,
            "shortuuid": self.shortuuid,
            "uuidv1": self.uuidv1,
            "mongodb_objectid": self.mongodb_objectid,
            "datetime": self.datetime,
            "datetimeRFC2822": self.datetimeRFC2822,
        }

    def allEncodingWithArgs(self) -> dict[str, EncodingType]:
        """Returns all encoding functions with required arguments."""

        return {"prefix_suffix": self.prefix_suffix}

    def availableEncoding(self) -> dict[str, EncodingType]:
        """Returns the encoding functions available to try out."""

        return {
            "base32": self.base32,
            "base64": self.base64,
            "urlencode": self.urlencode,
            "hexint": self.hexint,
            "hexstr": self.hexstr,
            "uniqid": self.uniqid,
            "shortuuid": self.shortuuid,
            "mongodb_objectid": self.mongodb_objectid,
            "uuidv1": self.uuidv1,
        }

    ### Encoding function list

    def prefix_suffix(
        self, token: str, encode: Optional[bool] = True, **kwargs: Any
    ) -> str:
        """Adds prefix and suffix if provided to a string value."""

        if encode:
            prefix = ""
            if "prefix" in kwargs and kwargs["prefix"]:
                prefix = kwargs["prefix"]
            suffix = ""
            if "suffix" in kwargs and kwargs["suffix"]:
                suffix = kwargs["suffix"]
            return f"{prefix}{token}{suffix}"
        raise ValueError("No possible to decode with prefix_suffix function.")

    def datetime(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from a timestamp to a date or vice versa"""

        if encode:
            timezone = kwargs["timezone"]
            date_format_of_token = kwargs["date_format_of_token"]
            return from_microsecond_timestamp(
                float(token),
                timezone=timezone,
                date_format_of_token=date_format_of_token,
            )
        return str(detect_datetime(token))

    def datetimeRFC2822(
        self, token: str, encode: Optional[bool] = True, **kwargs: Any
    ) -> str:
        """Converts from a string RFC2822 datetime to a date or vice versa"""

        if encode:
            timezone = kwargs["timezone"]
            date = from_microsecond_timestamp(
                float(token),
                timezone=timezone,
                date_format_of_token="%a, %d %b %Y %H:%M:%S %Z",
            )
            # No existing method with datetime to get the RFC2822, so we go with regex :
            # "UTC(+|-)07:00" -> "(+|-)0700"
            date = re.sub(r"UTC(.)(\d{2}):(\d{2})", r"\1\2\3", date)
            # "UTC" -> "+0000"
            date = re.sub(r"UTC$", r"+0000", date)
            return date
        return str(detect_datetime(token))

    def timestamp(
        self, token: str, encode: Optional[bool] = True, **kwargs: Any
    ) -> str:
        """Check if it's a timestamp value"""

        if encode:
            return to_microsecond_timestamp(token)
        return from_microsecond_timestamp(float(token))

    def mongodb_objectid(
        self, token: str, encode: Optional[bool] = True, **kwargs: Any
    ) -> str:
        """Converts from an PostgreSQL to a timestamp or vice versa"""

        if encode:
            init_token = kwargs["init_token"]
            return to_mongodb_objectid(token, init_token)
        return from_mongodb_objectid(token)

    def uuidv1(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from an UUID to a timestamp or vice versa"""

        if encode:
            init_token = kwargs["init_token"]
            return to_uuidv1(token, init_token)
        return from_uuidv1(token)

    def base32(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from a base32-encoded value to a string or vice versa"""

        if encode:
            return base64.b32encode(str(token).encode()).decode()
        return base64.b32decode(str(token).encode()).decode()

    def base64(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from a base64-encoded value to a string or vice versa"""

        if encode:
            return base64.b64encode(str(token).encode()).decode()
        return base64.b64decode(str(token).encode()).decode()

    def urlencode(
        self, token: str, encode: Optional[bool] = True, **kwargs: Any
    ) -> str:
        """Converts from a URL-encoded value to a string or vice versa"""

        if encode:
            return urlencode(token)
        return urldecode(token)

    def shortuuid(
        self, token: str, encode: Optional[bool] = True, **kwargs: Any
    ) -> str:
        """Converts from a shortuuid value to a UUID or vice versa"""

        if encode:
            return shortuuid.encode(uuid.UUID(str(token)))
        return str(shortuuid.decode(token))

    def uniqid(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from a uniqid value to a timestamp or vice versa"""

        if encode:
            return uniqid(float(token))
        return deuniqid(token)

    def hexint(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from a int-hex value to a timestamp or vice versa"""

        if encode:
            return hex(int(float(token))).replace("0x", "")
        return str(int(token, 16))

    def hexstr(self, token: str, encode: Optional[bool] = True, **kwargs: Any) -> str:
        """Converts from a str-hex value to a timestamp or vice versa"""

        if encode:
            return str(token).encode().hex()
        return bytes.fromhex(str(token)).decode()

    # Hashing

    def allHashIdentifiers(self) -> dict[str, HashIdentifierType]:
        """Returns all hash identifier functions."""

        return {
            "md5": self.is_md5,
            "sha1": self.is_sha1,
            "sha224": self.is_sha224,
            "sha256": self.is_sha256,
            "sha384": self.is_sha384,
            "sha512": self.is_sha512,
            "sha3_224": self.is_sha3_224,
            "sha3_256": self.is_sha3_256,
            "sha3_384": self.is_sha3_384,
            "sha3_512": self.is_sha3_512,
            "blake_256": self.is_blake_256,
            "blake_512": self.is_blake_512,
        }

    def allHashing(self) -> dict[str, HashingType]:
        """Returns all hash functions in dict with name as key."""

        return {
            "md5": self.md5,
            "sha1": self.sha1,
            "sha224": self.sha224,
            "sha256": self.sha256,
            "sha384": self.sha384,
            "sha512": self.sha512,
            "sha3_224": self.sha3_224,
            "sha3_256": self.sha3_256,
            "sha3_384": self.sha3_384,
            "sha3_512": self.sha3_512,
            "blake_256": self.blake_256,
            "blake_512": self.blake_512,
        }

    def intHashing(self) -> dict[str, HashingType]:
        """Returns all hash functions for a integer timestamp."""

        return self.allHashing()

    def floatHashing(self) -> dict[str, HashingType]:
        """Returns all hash functions for a integer timestamp."""

        return self.allHashing()

    def availableHashingDict(
        self, token: Optional[str] = None
    ) -> dict[str, HashingType]:
        """Returns the hash functions available by identifying the input hash."""

        if not token:
            return self.allHashing()
        hashes: dict[str, HashingType] = {}
        identifiers = self.allHashIdentifiers()
        for h in identifiers:
            is_hash = identifiers[h]
            hash = is_hash(str(token))
            if hash != None:
                hashes[h] = hash
        return hashes

    def hashing_with_prefix(
        self,
        hash_func: Callable[[str], str],
        prefixes: list[str],
        suffixes: list[str],
        token: str,
        values: tuple[str, str],
    ) -> tuple[Optional[str], Optional[str], Optional[str]]:
        """Returns timestamp and encoded value if hashed value from various provided prefixes and suffixes"""

        encoded_timestamp, timestamp = values
        if token == hash_func(encoded_timestamp):
            return timestamp, None, None
        for prefix in prefixes:
            value = "%s%s" % (prefix, encoded_timestamp)
            if token == hash_func(value):
                return timestamp, prefix, None
        for suffix in suffixes:
            value = "%s%s" % (encoded_timestamp, suffix)
            if token == hash_func(value):
                return timestamp, None, suffix
        for prefix in prefixes:
            for suffix in suffixes:
                value = "%s%s%s" % (prefix, encoded_timestamp, suffix)
                if token == hash_func(value):
                    return timestamp, prefix, suffix
        return None, None, None

    def _multithread_decrypt(
        self,
        token: str,
        possibleTokens: GeneratorLen,
        hash_func: Callable[[str], str],
        prefixes: list[str],
        suffixes: list[str],
        nb_threads: int = DEFAULT_THREAD_NUMBER,
    ) -> tuple[Optional[str], Optional[str], Optional[str]]:
        """Decrypts a timestamp-based value by using multithreading"""

        data = None, None, None
        chunksize = round(pow(len(possibleTokens) / nb_threads, 0.5))
        with concurrent.futures.ProcessPoolExecutor(max_workers=nb_threads) as executor:
            for timestamp, prefix, suffix in executor.map(
                partial(self.hashing_with_prefix, hash_func, prefixes, suffixes, token),
                possibleTokens,
                chunksize=chunksize,
            ):
                if timestamp:
                    data = timestamp, prefix, suffix
                    break
        return data

    def _native_decrypt(
        self,
        token: str,
        possibleTokens: GeneratorLen,
        hash_func: Callable[[str], str],
        prefixes: list[str],
        suffixes: list[str],
    ) -> tuple[Optional[str], Optional[str], Optional[str]]:
        """Decrypts a timestamp-based value by using naive method"""

        for value, timestamp in possibleTokens:
            timestamp, prefix, suffix = self.hashing_with_prefix(
                hash_func, prefixes, suffixes, token, (value, timestamp)
            )
            if timestamp:
                return timestamp, prefix, suffix
        raise NotAHash(f"It is not a hash")

    def is_hash(self, hash: str, hash_func: HashingType) -> Optional[HashingType]:
        """Generic function to detect a hash value"""

        if (
            hash.isdigit() == False
            and hash.isalpha() == False
            and hash.isalnum() == True
            and len(hash) == len(hash_func("FIXED_VALUE", encode=True))
        ):
            return hash_func
        return None

    def hash(
        self,
        hash_func: Callable[[str], str],
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        """Generic function for hashing or decrypting a value based on a timestamp."""

        if encode:
            return hash_func(token)
        if not possibleTokens:
            raise ValueError("No possibleTokens")
        prefixes: list[str] = (
            kwargs["prefixes"]
            if "prefixes" in kwargs and isinstance(kwargs["prefixes"], list)
            else []
        )
        suffixes: list[str] = (
            kwargs["suffixes"]
            if "suffixes" in kwargs and isinstance(kwargs["suffixes"], list)
            else []
        )
        if multithreading and multithreading > 1:
            return self._multithread_decrypt(
                token,
                possibleTokens,
                hash_func,
                prefixes,
                suffixes,
                nb_threads=multithreading,
            )
        return self._native_decrypt(
            token,
            possibleTokens,
            hash_func,
            prefixes,
            suffixes,
        )

    ### Hash function list

    def is_md5(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.md5)

    def md5(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.md5,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha1(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha1)

    def sha1(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha1,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha224(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha224)

    def sha224(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha224,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha256(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha256)

    def sha256(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha256,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha384(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha384)

    def sha384(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha384,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha512(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha512)

    def sha512(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha512,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha3_224(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha3_224)

    def sha3_224(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha3_224,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha3_256(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha3_256)

    def sha3_256(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha3_256,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha3_384(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha3_384)

    def sha3_384(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha3_384,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_sha3_512(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.sha3_512)

    def sha3_512(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.sha3_512,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_blake_256(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.blake_256)

    def blake_256(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.blake_256,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )

    def is_blake_512(self, hash: str) -> Optional[HashingType]:
        return self.is_hash(hash, self.blake_512)

    def blake_512(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        return self.hash(
            Hashes.blake_512,
            token,
            possibleTokens=possibleTokens,
            encode=encode,
            multithreading=multithreading,
            **kwargs,
        )



# -*- coding: utf-8 -*-
# Author: Aethlios <tom.chambaretaud@protonmail.com>

from os import cpu_count, path

OUTPUT_STRINGS = {
    "ROLEPLAY": {
        "TIMESTAMP_FOUND": "We found my precious! {timestamp} (prefix: {prefix} / suffix: {suffix})",
        "TOKEN_FORMAT": 'The way to get my precious is "{formats}"',
        "OUTPUT_IN_FILE": 'I\'ve hidden my precious ones in this dark corner: "{output}"',
        "INT_TIMESTAMP": "Please set you precious with --only-int-timestamp",
        "FAIL": "Stupid fat hobbit!",
    },
    "NOJOKE": {
        "TIMESTAMP_FOUND": "The token may be based on a timestamp: {timestamp} (prefix: {prefix} / suffix: {suffix})",
        "TOKEN_FORMAT": 'The convertion logic is "{formats}"',
        "OUTPUT_IN_FILE": 'Tokens have been exported in "{output}"',
        "INT_TIMESTAMP": "Please set the optional argument --only-int-timestamp",
        "FAIL": "It's not timestamp-based.",
    },
}

PROG_DESCRIPTION = "Reset Tolkien can be used to find out whether a provided token is based on a timestamp, from a timestamp corresponding to the period in which it was generated."

NB_DAYS_LIMIT = 365 * 1
TIMESTAMP_HEX_LENGTH = 8
TIMESTAMP_STR_LENGTH = 10
DEFAULT_DECIMAL_LENGTH = 7
UUID_DECIMAL_LENGTH = 7
DEFAULT_TIMERANGE_FOR_INT_TIMESTAMP = 60
DEFAULT_TIMERANGE_FOR_FLOAT_TIMESTAMP = 2
MIN_LENGTH_OF_FLOATING_TIMESTAMP_HEX = 8
MAX_LENGTH_OF_FLOATING_TIMESTAMP_HEX = 13

cpu = cpu_count()
DEFAULT_THREAD_NUMBER = cpu if cpu else 1
MIN_DEPTH_LEVEL = 1
MAX_DEPTH_LEVEL = 3
DEFAULT_DEPTH_LEVEL = MAX_DEPTH_LEVEL
MAX_VERBOSITY_LEVEL = 2
DEFAULT_VERBOSITY_LEVEL = 0

DEFAULT_CONFIG_FILENAME = "default.yml"
DEFAULT_CONFIG_FILE = path.join(path.dirname(path.realpath(__file__)), "config", DEFAULT_CONFIG_FILENAME)

SERVER_DATE_FORMAT = "%a, %d %b %Y %H:%M:%S %Z"

PARTIAL_FUNC_NAME_REGEX = r"functools\.partial\(<bound method Formatter\.([a-zA-Z_]*) of <src\.format\.Formatter object at [0-9a-fx]*>>, (.*)\)"

# 12/12/2023 or 5/5/2023 or 12/12/23 or 2023/12/12
DATE_SLASH_REGEX = r"[0-9]{1,4}\/[0-9]{1,2}\/[0-9]{1,4}"
# 12-12-2023 or 5-5-2023 or 12-12-23 or 2023-12-12
DATE_DASH_REGEX = r"[0-9]{1,4}\-[0-9]{1,2}\-[0-9]{1,4}"
# 22:32:47
TIME_REGEX = r"(?:\d{2}):(?:\d{2}):(?:\d{2})"
# January or Jan
MONTH_REGEX = r"(?:J(anuary|u(ne|ly))|February|Ma(rch|y)|A(pril|ugust)|(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)|(September|October|November|December))"



# __init__.py

__version__ = version = "1.1.1"
__version_tuple__ = version_tuple = (1, 1, 1)
__copyright__ = "Copyright (c) 2024 Tom CHAMBARETAUD"

from resetTolkien.cli import main

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Author: Aethlios <tom.chambaretaud@protonmail.com>

import uuid
import math
import random
import datetime
import argparse
import ast

from yaml import safe_load
from typing import Generator, Callable, Any, Optional, TypeAlias, Protocol
from typing_extensions import Self
from functools import partial
from urllib.parse import unquote_plus

import dateparser  # type: ignore
import hidateinfer  # type: ignore

from resetTolkien.constants import (
    NB_DAYS_LIMIT,
    TIMESTAMP_HEX_LENGTH,
    MIN_DEPTH_LEVEL,
    MAX_DEPTH_LEVEL,
    MIN_LENGTH_OF_FLOATING_TIMESTAMP_HEX,
    MAX_LENGTH_OF_FLOATING_TIMESTAMP_HEX,
    SERVER_DATE_FORMAT,
    UUID_DECIMAL_LENGTH,
)


# EXCEPTION


class NotAHash(Exception):
    pass


# Custom Float


def remove_last_zero(s: str) -> str:
    """Removes the 0 at the end of a decimal value"""

    if len(s) == 0:
        return "0"
    if s[-1] == "0":
        return remove_last_zero(s[:-1])
    return s


class CustomFloat:
    """Non-optimized but useful class for manipulating floats
    without rounding problems on decimal values."""

    def __init__(self, value: float, decimal_length: int):
        self.decimal_length = decimal_length
        self.value = round(value * 10**self.decimal_length)

    def integer(self):
        return self.value // 10**self.decimal_length

    def decimal(self):
        return self.value - self.integer() * 10**self.decimal_length

    def __repr__(self):
        return f"{self.integer()}.{remove_last_zero(str(self.decimal()))}"


# GENERATOR

TimestampGenerator: TypeAlias = Generator[tuple[str, str], None, None]


class AlternativeGen:
    """Generic class for defining the generator by alternative trial
    [0, 1, -1, 2, -2, ...]"""

    def __init__(self, range: int) -> None:
        self.range = range
        self.flipper = True
        self.n = 0

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> int:
        return self.next()

    def next(self) -> int:
        if self.n > self.range - 1:
            raise StopIteration()

        self.n += 1
        self.flipper = not self.flipper

        return (-1 if self.flipper else 1) * math.floor(self.n / 2)


class GeneratorLen(object):
    """Generic class for defining the length of a finite generator"""

    def __init__(self, gen: TimestampGenerator, length: int) -> None:
        self.gen = gen
        self.length = length

    def __type__(self) -> type:
        return type(self.gen)

    def __len__(self) -> int:
        return self.length

    def __iter__(self) -> TimestampGenerator:
        return self.gen


# TYPE


class SplitArgs(argparse.Action):
    """Action class to parse str with delimiter to list"""

    delimiter = ","

    def __call__(self, parser, namespace, values, option_string=None):  # type: ignore
        setattr(namespace, self.dest, [value.strip() for value in values.split(self.delimiter)])  # type: ignore


class EncodingType(Protocol):
    """Type for Encoding functions"""

    def __call__(self, token: str, encode: bool = True, **kwargs: Any) -> str:
        ...

    def __name__(self) -> str:
        return self.__call__.__name__


class HashingType(Protocol):
    """Type for Hashing functions"""

    def __call__(
        self,
        token: str,
        possibleTokens: Optional[GeneratorLen] = None,
        encode: Optional[bool] = True,
        multithreading: Optional[int] = None,
        **kwargs: Any,
    ) -> str | tuple[Optional[str], Optional[str], Optional[str]]:
        ...

    def __name__(self) -> str:
        return self.__call__.__name__


FormatType: TypeAlias = EncodingType | HashingType | partial[Any]
HashIdentifierType: TypeAlias = Callable[[str], Optional[HashingType]]


# Function

# Credit to https://github.com/intruder-io/guidtool
def uuid1(node: int, clock_seq: int, timestamp: float):
    """Generates a UUID from a host ID, sequence number, and the current time.
    If 'node' is not given, getnode() is used to obtain the hardware
    address.  If 'clock_seq' is given, it is used as the sequence number;
    otherwise a random 14-bit sequence number is chosen."""

    if float(timestamp) == int(float(timestamp)):
        timestamp = int(timestamp)

    if type(timestamp) == int:
        timestamp = timestamp * 10**UUID_DECIMAL_LENGTH
    else:
        splitted = str(timestamp).split(".")
        splitted[1] += "0" * (UUID_DECIMAL_LENGTH - len(splitted[1]))
        timestamp = int("".join(splitted))

    timestamp = timestamp + 0x01B21DD213814000
    time_low = timestamp & 0xFFFFFFFF
    time_mid = (timestamp >> 32) & 0xFFFF
    time_hi_version = (timestamp >> 48) & 0x0FFF
    clock_seq_low = clock_seq & 0xFF
    clock_seq_hi_variant = (clock_seq >> 8) & 0x3F
    return uuid.UUID(
        fields=(
            time_low,
            time_mid,
            time_hi_version,
            clock_seq_hi_variant,
            clock_seq_low,
            node,
        ),
        version=1,
    )


def to_uuidv1(timestamp: str, init_token: str):
    """Converts a timestamp to a UUID via a provided UUID"""

    try:
        u = uuid.UUID(init_token)
    except (ValueError, AttributeError):
        raise ValueError("Not a UUID")
    if u.version != 1:
        raise ValueError("Not a UUIDv1")
    return str(uuid1(u.node, u.clock_seq, float(timestamp)))


def from_uuidv1(token: str):
    """Extracts a timestamp from an UUID"""

    try:
        u = uuid.UUID(token)
    except (ValueError, AttributeError):
        raise ValueError("Not a UUID")
    if u.version != 1:
        raise ValueError("Not a UUIDv1")
    return str((u.time - 122192928000000000) / 10000000)


class MongoDBObjectID:
    def __init__(self, s: str) -> None:
        if len(s) != 24:
            raise ValueError("Wrong length")
        self.timestamp = int(s[:8], 16)
        self.machine = int(s[8:14], 16)
        self.process = int(s[14:18], 16)
        self.counter = int(s[18:24], 16)

    def set_timestamp(self, timestamp: str):
        self.timestamp = int(timestamp)

    def __str__(self) -> str:
        return "%08x%06x%02x%06x" % (
            self.timestamp,
            self.machine,
            self.process,
            self.counter,
        )


def to_mongodb_objectid(timestamp: str, init_token: str):
    try:
        u = MongoDBObjectID(init_token)
    except (ValueError, AttributeError):
        raise ValueError("Not a MongoDBObjectID")
    u.set_timestamp(timestamp)
    return str(u)


def from_mongodb_objectid(token: str):
    try:
        u = MongoDBObjectID(token)
    except (ValueError, AttributeError):
        raise ValueError("Not a MongoDBObjectID")
    return str(u.timestamp)


# Credit to https://github.com/Riamse/python-uniqid
def uniqid(timestamp: float, prefix: str = "", more_entropy: bool = False):
    """uniqid([prefix=''[, more_entropy=False]]) -> str
    Gets a prefixed unique identifier based on the current
    time in microseconds.
    prefix
        Can be useful, for instance, if you generate identifiers
        simultaneously on several hosts that might happen to generate
        the identifier at the same microsecond.
        With an empty prefix, the returned string will be 13 characters
        long. If more_entropy is True, it will be 23 characters.
    more_entropy
        If set to True, uniqid() will add additional entropy (using
        the combined linear congruential generator) at the end of
        the return value, which increases the likelihood that
        the result will be unique.
    Returns the unique identifier, as a string."""

    # m = time.time()
    m = timestamp

    sec = math.floor(m)
    usec = round(1000000 * (m - sec))
    if more_entropy:
        lcg = random.random()
        the_uniqid = "%08x%05x%.8F" % (sec, usec, lcg * 10)
    else:
        the_uniqid = "%8x%05x" % (sec, usec)

    the_uniqid = prefix + the_uniqid
    return the_uniqid


def deuniqid(value: str) -> str:
    """Converts the timestamp from an uniqid value"""

    if (
        len(value) < MIN_LENGTH_OF_FLOATING_TIMESTAMP_HEX
        or len(value) > MAX_LENGTH_OF_FLOATING_TIMESTAMP_HEX
    ):
        raise ValueError("The input is too short")
    return (
        str(int(value[:TIMESTAMP_HEX_LENGTH], 16))
        + "."
        + str(int(value[TIMESTAMP_HEX_LENGTH:], 16))
    )


def urlencode(value: str) -> str:
    """URL-encodes all values"""

    return "".join("%{0:0>2x}".format(ord(char)) for char in str(value))


def urldecode(value: str) -> str:
    """URL-decodes all URL-encoded values - trigger an exception if no value is encoded"""

    decoded = unquote_plus(value)
    if decoded == value:
        raise ValueError("data is not urlencoded")
    return decoded


def from_microsecond_timestamp(
    token: float, timezone: int = 0, date_format_of_token: str | None = None
) -> str:
    """Converts a timestamp to a standard string of datetime or formated by date_format_of_token"""

    timezone_delta = datetime.timedelta(hours=timezone)
    d = (
        datetime.datetime.utcfromtimestamp(token).replace(
            tzinfo=datetime.timezone(timezone_delta)
        )
        + timezone_delta
    )
    now = datetime.datetime.now(datetime.timezone(datetime.timedelta(hours=timezone)))
    if d < now - datetime.timedelta(days=NB_DAYS_LIMIT) or d > now + datetime.timedelta(
        days=NB_DAYS_LIMIT
    ):
        raise ValueError("Not a timestamp")
    if date_format_of_token:
        return d.strftime(date_format_of_token)
    return d.isoformat()


def to_microsecond_timestamp(d: str) -> str:
    """Converts a standard string of datetime to a timestamp"""

    return str(datetime.datetime.fromisoformat(d).timestamp())


def detect_datetime(token: str) -> datetime.datetime:
    """Extracts a date from a string - trigger an exception if it is not a date"""

    date = dateparser.parse(token)
    if not date:
        raise ValueError("Not a date")
    return date


def possible_date_format_of_token(token: str) -> Any:
    """Tries to extract possible datetime formats from a datetime string"""

    return hidateinfer.infer([token])  # type: ignore


# FORMAT


def import_format_with_args(
    format: str, formats: dict[str, EncodingType]
) -> tuple[FormatType, dict[str, Any]]:
    """Function to import the format function with args"""

    class FunctionNamePrinter(ast.NodeVisitor):
        def __init__(self, formats: dict[str, EncodingType]) -> None:
            self.formats: dict[str, EncodingType] = formats
            super().__init__()

        def visit_Module(self, node: ast.Module) -> Any:
            if (
                len(node.body) > 0
                and isinstance(node.body[0], ast.Expr)
                and isinstance(node.body[0].value, ast.Call)
            ):
                return self.visit_Call(node.body[0].value)
            raise ValueError(f"No good expression : {node.body}")

        def visit_Call(self, node: ast.Call) -> Any:
            name = None
            if isinstance(node.func, ast.Name):
                name = self.visit_Name(node.func)
            if not name:
                raise ValueError(f"No name : {node.func}")
            args: dict[str, str] = {}
            for keyword in node.keywords:
                k: dict[str, str] = self.visit_keyword(keyword)
                args.update(k)
            return name, args

        def visit_Name(self, node: ast.Name) -> Any:
            func_name: str = str(node.id)
            for f in self.formats:
                if func_name == f:
                    return self.formats[f]
            raise ValueError(f"{node.id} doesn't exist")

        def visit_Constant(self, node: ast.Constant) -> Any:
            return node.value

        def visit_keyword(self, node: ast.keyword) -> Any:
            if isinstance(node.value, ast.Constant):
                return {node.arg: self.visit_Constant(node.value)}
            raise ValueError(f"No constant : {node.value}")

    module: ast.Module = ast.parse(format)
    visitor = FunctionNamePrinter(formats)
    return visitor.visit(module)


# TimestampHashFormat


class TimestampHashFormat:
    """Class used to define the timestamp format parameters
    provided to decrypt a hash."""

    def __init__(
        self,
        description: str,
        hashes_by_type: dict[str, HashingType],
        timestamp_type_func: Callable[[str], Any],
        range_limit: int,
        level: int = MIN_DEPTH_LEVEL,
        formats_output: Optional[list[FormatType]] = None,
    ) -> None:
        self.description = description
        self.hashes_by_type = hashes_by_type
        self.timestamp_type_func = timestamp_type_func
        self.range_limit = range_limit
        self.level = level
        self.formats_output: list[FormatType] = (
            [] if formats_output is None else formats_output
        )

    def __repr__(self) -> str:
        return f"({self.level}) {self.description}"


def import_from_yaml(
    filename: str,
    selected_level: int,
    int_range_limit: int,
    float_range_limit: int,
    intHashing: dict[str, HashingType],
    floatHashing: dict[str, HashingType],
    allEncoding: dict[str, EncodingType],
) -> list[TimestampHashFormat]:
    """From a YAML file, imports a TimestampHashFormat list"""

    r: list[TimestampHashFormat] = []
    with open(filename) as f:
        hashFormats = safe_load(f)
        for f in hashFormats:
            hashFormat = hashFormats[f]
            level = hashFormat["level"] if "level" in hashFormat else MIN_DEPTH_LEVEL
            if level > MAX_DEPTH_LEVEL:
                raise ValueError(f"Please set a value < {MAX_DEPTH_LEVEL}")
            if selected_level < level:
                continue
            description = (
                hashFormat["description"]
                if "description" in hashFormat
                else "Not described"
            )
            if hashFormat["timestamp_type"] == "int":
                hashes_by_type = intHashing.copy()
                timestamp_type_func: Callable[[str], float] = lambda timestamp: int(
                    float(timestamp)
                )
                range_limit = int_range_limit
            elif hashFormat["timestamp_type"] == "float":
                hashes_by_type = floatHashing.copy()
                timestamp_type_func: Callable[[str], float] = lambda timestamp: float(
                    timestamp
                )
                range_limit = float_range_limit
            else:
                raise ValueError('Please set "int" or "float" types.')
            formats: list[FormatType] = []
            if "formats" in hashFormat:
                for f in hashFormat["formats"]:
                    if not f in allEncoding.keys():
                        raise ValueError(
                            f"Please set a valid format : {allEncoding.keys()}."
                        )
                    else:
                        formats.append(allEncoding[f])
            r.append(
                TimestampHashFormat(
                    description=description,
                    hashes_by_type=hashes_by_type,
                    timestamp_type_func=timestamp_type_func,
                    range_limit=range_limit,
                    level=level,
                    formats_output=formats,
                )
            )
    return sorted(r, key=lambda x: x.level)


# EXAMPLE VALUE

server_date_example = datetime.datetime.now(datetime.timezone.utc).strftime(
    SERVER_DATE_FORMAT
)



# -*- coding: utf-8 -*-
# Author: Aethlios <tom.chambaretaud@protonmail.com>

import datetime
import base64
from binascii import hexlify
import hashlib
import uuid
import time

from resetTolkien.resetTolkien import ResetTolkien, FormatType
from resetTolkien.utils import uuid1, uniqid, urlencode, MongoDBObjectID
from resetTolkien.constants import UUID_DECIMAL_LENGTH

import sys

verbosity = 0
if len(sys.argv) == 3 and sys.argv[1] == "-v":
    verbosity = int(sys.argv[2])

threads = 8

TIMEDELTA_WITH_FLOAT_VALUE = 1.1
TIMEDELTA_WITH_INT_VALUE = 30

OK = "\033[92mOK\033[0m"
NOK = "\033[91mNOK\033[0m"


def benchmark_multithread() -> None:
    timestamp_input = datetime.datetime.now().timestamp()
    token = hashlib.md5("value".encode()).hexdigest()
    print("native")
    tolkien = ResetTolkien(
        token=token,
        verbosity=2,
    )
    t = time.time()
    result = tolkien.detectFormat(timestamp=timestamp_input, nb_threads=1)
    t2 = time.time()
    print(result)
    print(f"Time : {t2-t}")
    print("multithread")
    tolkien = ResetTolkien(
        token=token,
        verbosity=2,
        prefixes=["a", "c"],
        suffixes=["b", "d"],
    )
    t = time.time()
    result = tolkien.detectFormat(timestamp=timestamp_input, nb_threads=8)
    t2 = time.time()
    print(result)
    print(f"Time : {t2-t}")


def _check(
    tolkien: ResetTolkien,
    value: str,
    init_token: str,
    results: list[tuple[tuple[str, str | None, str | None], list[FormatType], bool]],
) -> tuple[bool, list[list[FormatType]]]:
    for _, formats, isBasedOnTimestamp in results:
        if isBasedOnTimestamp:
            try:
                result = tolkien.encode(value, formats=formats)
                if result != init_token and len(results) > 0:
                    print(formats)
                    print(f"result : {result} - expected result : {init_token} - value : {value}")
                return (
                    result == init_token,
                    [formats for _, formats, _ in results if isBasedOnTimestamp],
                )
            except:
                pass
    return (False, [])


def check(
    value: str | float | int,
    init_token: str,
    description: str = "Generic test",
    decimal_length: int = 6,
    timestamp_input: float | None = None,
    prefixes: list[str] | None = None,
    suffixes: list[str] | None = None,
    date_format_of_token: str | None = None,
    timezone: int = 0,
    force_success: bool = False,
) -> None:
    prefixes = prefixes if prefixes else []
    suffixes = suffixes if suffixes else []

    if verbosity >= 1:
        print(f"Value : {value}")
        print(f"Token : {init_token}")

    if isinstance(value, int) or isinstance(value, float):
        value = str(value)

    tolkien = ResetTolkien(
        token=init_token,
        timezone=timezone,
        verbosity=verbosity,
        prefixes=prefixes,
        suffixes=suffixes,
        date_format_of_token=date_format_of_token,
        decimal_length=decimal_length,
    )
    start = time.time()
    results = tolkien.detectFormat(timestamp=timestamp_input, nb_threads=threads)
    if not results:
        raise Exception("No defined token")
    end = time.time()
    if verbosity >= 1:
        print(f"Partial results : {results}")
    success, possible_formats = _check(tolkien, value, init_token, results)
    print(
        f"[{round(end - start, 3)}s] {description} : {(OK if success or force_success else NOK)} (possibles formats : {len(possible_formats)})"
    )

print("[+] C=heck formats")

timestamp = int(datetime.datetime.now().timestamp())
token = base64.b32encode(str(timestamp).encode()).decode()
check(timestamp, token, description="Int timestamp with base32")

timestamp = int(datetime.datetime.now().timestamp())
token = base64.b64encode(str(timestamp).encode()).decode()
check(timestamp, token, description="Int timestamp with base64")

timestamp = int(datetime.datetime.now().timestamp())
token = urlencode(str(timestamp))
check(timestamp, token, description="Int timestamp with urlencode")

timestamp = int(datetime.datetime.now().timestamp())
token = hex(timestamp)[2:]
check(timestamp, token, description="Int timestamp with hexint")

timestamp = int(datetime.datetime.now().timestamp())
token = hexlify(str(timestamp).encode()).decode()
check(timestamp, token, description="Int timestamp with hexstr")

timestamp = int(datetime.datetime.now().timestamp())
token = str(uniqid(timestamp))
check(
    timestamp,
    token,
    description="Int timestamp with uniqid",
)

timestamp = int(datetime.datetime.now().timestamp())
mongoOID_example = "65b23087d5888f1392d74c95"
u = MongoDBObjectID(mongoOID_example)
u.set_timestamp(timestamp)
token = str(u)
check(timestamp, token, description="Int timestamp with mongodb_objectid")

timestamp = int(datetime.datetime.now().timestamp())
u = uuid.uuid1()
token = str(uuid1(u.node, u.clock_seq, timestamp))
check(timestamp, token, description="Int timestamp with uuidv1", decimal_length=UUID_DECIMAL_LENGTH)

timestamp = int(datetime.datetime.now().timestamp())
token = base64.b64encode(hex(timestamp)[2:].encode()).decode()
check(timestamp, token, description="Int timestamp with hexint and base64")

timestamp = int(datetime.datetime.now().timestamp())
token = base64.b64encode(hexlify(str(timestamp).encode())).decode()
check(timestamp, token, description="Int timestamp with hexstr and base64")

timestamp = datetime.datetime.now().timestamp()
token = base64.b64encode(uniqid(timestamp).encode()).decode()
check(timestamp, token, description="Float timestamp with uniqid and base64")

timestamp = datetime.datetime.now().timestamp()
token = hexlify(str(timestamp).encode()).decode()
check(timestamp, token, description="Float timestamp with hexstr")

timestamp = datetime.datetime.now().timestamp()
token = base64.b64encode(hexlify(str(timestamp).encode())).decode()
check(timestamp, token, description="Float timestamp with hexstr and base64")

timestamp = datetime.datetime.now().timestamp()
token = uniqid(timestamp)
check(timestamp, token, description="Float timestamp with uniqid")

print("[+] Other checks")

timestamp = int(datetime.datetime.now().timestamp())
token = base64.b64encode(
    hashlib.sha1(str(timestamp).encode()).hexdigest().encode()
).decode()

tolkien = ResetTolkien(token=token, formats=["base64", "sha1"])
success = token == tolkien.encode(str(timestamp))
print(f"Format importation : {(OK if success else NOK)}")

tokens = list(tolkien.generate_possible_token(timestamp, range_limit=4))
success = len(tokens) == 4 and len(set(tokens)) == len(tokens) and tokens[0][0] == token
print(f"Possible token exportation : {(OK if success else NOK)}")

print("[+] Check prefix/suffix")

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
value = str(timestamp)
token = hashlib.md5(value.encode()).hexdigest()
check(
    timestamp,
    token,
    description="Check prefix/suffix from native value with md5",
    timestamp_input=timestamp_input,
    prefixes=["hello", "you"],
    suffixes=["1", "2"],
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
value = "%s%s" % ("you", str(timestamp))
token = hashlib.md5(value.encode()).hexdigest()
check(
    timestamp,
    token,
    description="Check prefix with md5",
    timestamp_input=timestamp_input,
    prefixes=["hello", "you"],
    suffixes=["1", "2"],
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
value = "%s%s" % (str(timestamp), "2")
token = hashlib.md5(value.encode()).hexdigest()
check(
    timestamp,
    token,
    description="Check suffix with md5",
    timestamp_input=timestamp_input,
    prefixes=["hello", "you"],
    suffixes=["1", "2"],
)

print("[+] Check datetime")

date = datetime.datetime.now(tz=datetime.timezone(datetime.timedelta(hours=-7)))
timestamp = date.timestamp()
timestamp_input = round(timestamp - 1, 6)
token = date.strftime("%a, %d %b %Y %H:%M:%S %Z")
check(
    timestamp,
    token,
    description="Datetime RFC2822 with timezone",
    timestamp_input=timestamp_input,
    date_format_of_token="%a, %d %b %Y %H:%M:%S %Z",
    timezone=-7,
)

date = datetime.datetime.now(tz=datetime.timezone.utc)
timestamp = date.timestamp()
timestamp_input = round(timestamp - 1, 6)
token = base64.b64encode(date.isoformat().encode()).decode()
check(
    timestamp,
    token,
    description="Datetime with base64",
    timestamp_input=timestamp_input,
    date_format_of_token="%Y-%m-%dT%H:%M:%S.%f+00:00",
)

print("[+] Check hashes")

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp + TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.md5(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with md5 for token after the timestamp",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = base64.b64encode(
    hashlib.md5(str(timestamp).encode()).hexdigest().encode()
).decode()
check(
    timestamp,
    token,
    description="Int timestamp with md5 and base64",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hexlify(hashlib.md5(str(timestamp).encode()).hexdigest().encode()).decode()
check(
    timestamp,
    token,
    description="Int timestamp with md5 and hexstr",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.md5(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with md5",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha1(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha1",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha224(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha224",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha256(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha256",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha384(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha384",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha512(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha512",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha3_224(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha3_224",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha3_256(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha3_256",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha3_384(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha3_384",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.sha3_512(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with sha3_512",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.blake2s(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with blake_256",
    timestamp_input=timestamp_input,
)

timestamp = int(datetime.datetime.now().timestamp())
timestamp_input = round(timestamp - TIMEDELTA_WITH_INT_VALUE, 6)
token = hashlib.blake2b(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Int timestamp with blake_512",
    timestamp_input=timestamp_input,
)

timestamp = datetime.datetime.now().timestamp()
timestamp_input = round(timestamp + TIMEDELTA_WITH_FLOAT_VALUE, 6)
u = uuid.uuid1()
token = str(uuid1(u.node, u.clock_seq, timestamp))
check(timestamp, token, description="Float timestamp with uuidv1", timestamp_input=timestamp_input)

timestamp = datetime.datetime.now().timestamp()
timestamp_input = round(timestamp + TIMEDELTA_WITH_FLOAT_VALUE, 6)
token = hashlib.blake2b(str(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Float timestamp with blake_512",
    timestamp_input=timestamp_input,
)

timestamp = datetime.datetime.now().timestamp()
timestamp_input = round(timestamp + TIMEDELTA_WITH_FLOAT_VALUE, 6)
token = hashlib.sha1(uniqid(timestamp).encode()).hexdigest()
check(
    timestamp,
    token,
    description="Uniquid timestamp with sha1",
    timestamp_input=timestamp_input,
)

timestamp = datetime.datetime.now().timestamp()
timestamp_input = round(timestamp - 1, 6)
value = "%s%s%s" % ("you", str(timestamp), "2")
token = hashlib.md5(value.encode()).hexdigest()
check(
    timestamp,
    token,
    description="Check prefix/suffix from float value with md5",
    timestamp_input=timestamp_input,
    prefixes=["hello", "you"],
    suffixes=["1", "2"],
)

date = datetime.datetime.now(tz=datetime.timezone.utc)
timestamp = date.timestamp()
timestamp_input = round(timestamp - 1, 6)
token = hashlib.md5(date.isoformat().encode()).hexdigest()
check(
    timestamp,
    token,
    description="Datetime with md5",
    timestamp_input=timestamp_input,
    date_format_of_token="%Y-%m-%dT%H:%M:%S.%f+00:00",
)

print("[+] Check not working")

timestamp = datetime.datetime.now().timestamp()
timestamp_input = round(timestamp + TIMEDELTA_WITH_FLOAT_VALUE, 6)
token = hashlib.sha3_512("fusbevuisbevuiesbvuiesbvsie".encode()).hexdigest()
check(
    timestamp,
    token,
    description="Unknown token with sha3_512",
    timestamp_input=timestamp_input,
    force_success=True,
)
